#!/usr/bin/env python
#-*- coding:utf-8 -*-

# Copyright (C) 2010 Antony Templier
# 
# This software is under WTFPL License:
# ------------------------------------------------------------------------------
#             DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#                     Version 2, December 2004
# 
#  Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
# 
#  Everyone is permitted to copy and distribute verbatim or modified
#  copies of this license document, and changing it is allowed as long
#  as the name is changed.
# 
#             DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#    TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
# 
#   0. You just DO WHAT THE FUCK YOU WANT TO.
# ------------------------------------------------------------------------------

"""
API for PmWiki page from both web site and local file.

Can read, write, and change attributes or encoding of a page.
Just use the PmwikiPage Class.
"""

__version__ = '0.1.0'
__author__ = "Antony Templier <antony.tempier@gmail.com>"
__date__ = "30 june 2011"
__license__ = "WTFPL"
__credits__ = ""

from sys import stderr
from os import path as p
from os import access, W_OK
from time import time
import re
import codecs
from StringIO import StringIO

from odict import odict
import logging
import urlparse
from urllib2 import Request, urlopen
from urllib import urlencode

import diff
from md5crypt import md5crypt, check_password, encrypt


#---------------------------------Handle Errors---------------------------------


class Error(Exception):
    def _say(self, msg):
        logging.error(msg)
        #stderr.write('{0}\n'.format(msg))

class ConnexionError(Error):
    def __init__(self, msg):
        self._msg = '[ Connexion Error ]: {0}'.format(msg)
        self._say(self._msg)

class FileAccessError(Error):
    def __init__(self, msg):
        self._msg = '[ File Access Error ]: {0}'.format(msg)
        self._say(self._msg)

class NoValidReachPointError(Error):
    def __init__(self, msg):
        self._msg = '[ No valid reach point Error ]: {0}'.format(msg)
        self._say(self._msg)

class AuthenticationError(Error):
    def __init__(self, msg):
        self._msg = '[ Authentication Error ]: {0}'.format(msg)
        self._say(self._msg)

class NonExistingPageError(Error):
    def __init__(self, msg):
        self._msg = '[ Non Existing Page Error ]: {0}'.format(msg)
        self._say(self._msg)

class BadNameError(Error):
    def __init__(self, msg):

        self._msg = '[ Bad Name Error ]: {0}'.format(msg)
        self._say(self._msg)

class BlockByWikiError(Error):
    def __init__(self, msg):

        self._msg = '[ Block By Wiki Error ]: {0}'.format(msg)
        self._say(self._msg)

class BadCharsetError(Error):
    def __init__(self, msg):

        self._msg = '[ Bad Charset Error ]: {0}'.format(msg)
        self._say(self._msg)


#-------------------------------------Tools-------------------------------------


def fmt_pageName(name):
    """
    Check if the name parameter is a good name for a Pmwiki page.
    If name isn't correct then we try to do conversion.
    If we can't return false

    For regexp, we search for someting like :
    
    allc = A-Z, a-z, 0-9, _, - and space
    nameof = [not starting with _,-,space] [allc] [not ending with _,-,space] 
    [nameof][ / or . ][nameof]  

    @type   name: String
    @param  name: the name of the page

    @rtype  String
    @rparam New name well formatted
    """

    reg =  r"""^(?!([-_]| ))[\w -]+(?<![-_])[/.](?![-_])[\w -]+(?<!([-_]| ))$"""
    s_name = re.compile(reg)
    r_name = s_name.search(name)
    if r_name is not None:

        result = r_name.group(0)
        result = result.replace('/','.')
        tmp = result.split('.')
        
        for i in range(0, len(tmp)):
            tmp[i] = ''.join([ ''.join([t[0].upper(), t[1:]]) \
            for t in tmp[i].split(' ') if t !=''])
        result = '.'.join( tmp )

        if result != name:
            logging.warning(
            "Page name '{0}' has been renamed to '{1}'".format(name, result))
        return result
    else:
        raise BadNameError("'{0}' is not a valid Pmwiki page name".format(name))

def src2txt(txt):
    """Convert pmwiki raw source to human readable source pmwiki syntax page """
    txt = txt.replace('%0a', '\n')
    txt = txt.replace('%3c', '<')
    txt = txt.replace('%25', '%')
    return txt

def txt2src(txt):
    """Convert human readable source pmwiki page to raw source"""
    txt = txt.replace('%', '%25')
    txt = txt.replace('\n','%0a')
    txt = txt.replace('<', '%3c')
    return txt

# decorators
def for_human(func):
    """Decorator for src2txt"""
    def do_convert(*arg, **args):
        return  src2txt( func(*arg, **args) )
    return do_convert


#-----------------------------------Class Page----------------------------------


class BasePage(object):
    """Base class for page"""
    def __init__(self, path, name):
        self._path = path
        self._name = name
        self._text = None
        self._charset = u'UTF-8'
    
    def set_charset(self, charset=None):
        """Set character encoding""" 
        if charset is None:
            raise BadCharsetError(
                'charset=??? is not set for {0}'.format(self._name)
            )
        try:
            codecs.lookup(charset)
        except LookupError, e:
            raise BadCharsetError(
                '{0} for {1}'.format(e, self._name)
            )
        else:
            self._charset = unicode(charset, charset)

    def check_path(self):
        """Check if the path to the page is valid"""
        raise NotImplementedError

    def read(self, author=None, password=None):
        """Read a page and return his content as Pmwiki syntax source text"""
        raise NotImplementedError

    def write(self, text=None, author=None, password=None):
        """Create or update a page from Pmwiki syntax source text"""
        raise NotImplementedError

    def change_attr(self, author=None, password=None, **attributes):
        """
        Change attributes'page to protect or unprotect the page with reading,
        writting, uploading and changing attributes. 
        """
        raise NotImplementedError

class HttpPage(BasePage):
    """A web PmWiki page: manipulation are done via http request."""

    def __init__(self, path, name):
        super(HttpPage, self).__init__(path, name)
        self._delete_word = 'delete'  

    def _try_openurl(self, source, params=None, headers=None):
        """
        Handle connexion for requested url

        @type   source: string
        @param  source: an url with action required. Action can be :
                        - edit
                        - source
                        - attr
                        Ex: http://www.pmwiki.org/?action=source

        @type   params: dict
        @param  params: Other parameters of the action

        @rtype  object urllib2.Request
        @rparam handler for the connected url
        """

        if params is not None:
            request = Request(source, urlencode(params))
        else:
            request = Request(source)

        if headers is not None:
            for h_name, h_value in headers.iteritems():
                request.add_header(h_name, h_value)

        try:
            handle_req = urlopen(request)
        except IOError, e:
            if hasattr(e, 'reason'):
                raise ConnexionError(
                   'We failed to reach a server: {0}'.format(e.reason)
                )
            elif hasattr(e, 'code'):
                if e.code == 404 and params is not None: 
                    if params['text'] == self._delete_word: 
                        logging.warning('{0} is deleted'.format(self._name))
                        return StringIO('page is deleted')
                else:
                    raise ConnexionError(
                        "The server couldn't fulfill the request: {0}".format(
                        e.code)
                    )
        else:
            return handle_req

    def check_path(self):
        """
        Test if we can reach the page's url

        @rtype    Boolean
        @rparam    True if we can reach a pmwiki site else raise an error
        """

        handle_page = self._try_openurl( self._path )
        text_page = handle_page.read()
        if not '<!--HTMLHeader-->' in text_page and \
           not '<!--PageText-->' in text_page:
            raise NoValidReachPointError(
                'Not a PmWiki (No <!--HTMLHeader--> and <!--PageText--> found)'
            )
        else:
            return True

    def read(self, author=None, password=None):
        """
        Read the page
        
        @type   author: string
        @param  author: author (as $authid) required when page is read protected
                        and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is read protected
        """

        url = '?'.join([self._path, 'action=source'])
        params={'n':self._name}

        if author is not None:
            params['author'] = author
        if password is not None:
            params['authid'] = author
            params['authpw'] = password

        result = self._try_openurl( url, params )
        text_page = result.read()
        #text_page = result

        if text_page[:9] == '<!DOCTYPE':
            message = 'trying to read {0}'.format(self._name)
            raise AuthenticationError(message)

        # There is no way to figure out te difference between an existing page
        # with no content and a real non existing page.
        # Until now, both cases raise an error ( also a page with no content is
        # not very useful).
        elif len(text_page) == 0: 
            message = '{0} has no content'.format(self._name)
            raise NonExistingPageError(message)

        #Catching charset
        headers = result.info()
        self.set_charset(headers.getparam('charset'))

        return unicode(text_page, self._charset)


    def write(self, text, author=None, password=None, charset=None):
        """
        Write the page. 
        
        Even if just charset is changed, a new entry history will be add.
        
        @type   author: string
        @param  author: author (as $authid) required when page is write 
                        protected and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is write protected

        @type   charset: String
        @param  charset: charset for character encoding, ex: UTF-8, ISO-8859-15
        """
        
        url = '?'.join([self._path, 'action=edit'])
        
        # Check if charset has changed
        if charset is not None and charset != self._charset:
            self.set_charset(charset)

        params={
            'n':self._name, 
            'post' : 1, 
            'text' : unicode(text).encode(self._charset)
        }

        if author is not None:
            params['author'] = author
        if password is not None:
            params['authid'] = author
            params['authpw'] = password


        headers = {
            #'Content-Type' : 'multipart/form-data',
            'Content-Type' : 'application/x-www-form-urlencoded',
            'Accept' : 'text/plain',
            'Accept-Charset' : '{0}'.format(self._charset)
        }

        result = self._try_openurl( url, params)#, headers )
        # I'ven't found another technique than checking source htlm 
        # to intercept these kinds of error
        rtxt = result.readlines()
       #rtxt = result.split('\n')
        for line in rtxt:

            # Need authentication when doing something
            if "<input type='password' name='authpw' class='inputbox' />" \
              in line:
                raise AuthenticationError('trying to write {0}'.format(self._name))
 
            # May append when need an author name 
            # when editing ($EnablePostAuthorRequired = 1) or
            # when blocklist prevent some post ($EnableBlocklist = 1) 
            if 'wikimessage' in line:
                line = line.strip('\n\t ')
                # We strip html tag ,
                msg = re.sub("< */? *.+? */?>",'',line)
                raise BlockByWikiError(''.join(msg))

    def change_attr(self, author=None, password=None, **attributes):
        """
        Change attributes'page to protect or unprotect the page with reading,
        writting, uploading and changing attributes.

        @type   author: string
        @param  author: author (as $authid) required when page is protected
                        and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is protected

        @type   attributes: dict
        @param  attribute: key elts in dict can be 'passwdedit', 'passwdread', 
                           'passwdupload' and 'passwdattr'
        """
        
        url = '?'.join([self._path, 'action=postattr'])
        params={'n':self._name}

        if author is not None:
            params['author'] = author
        if password is not None:
            params['authid'] = author
            params['authpw'] = password

        params.update(attributes)

        result = self._try_openurl( url, params )

class FilePage(BasePage):
    """A Pmwiki page as a local file"""

    def __init__(self, path, name):
        super(FilePage, self).__init__(path, name)
        self._page_elts = None

    def _try_fileaccess(self, path, mode):
        """
        Check for access on local file sytem

        @type   path: String
        @param  path: path of the file

        @type   mode: String
        @param  mode: Set type of access. 'r' -> read, 'w' -> write ...

        @rtype  Object
        @rparam file handler
        """
        try:
            handler = open(path, mode)
        except IOError, e:
            raise FileAccessError( e )
        else:
            return handler

    def _parse_file(self, source):
        """
        Fill page_elts dict from a Pmwiki page file.
        Change charset if found.  

        @type   source: String
        @param  source: raw source of a pmwikipage file
        """
        elts = odict()
        
        if not isinstance(source, unicode):
            # Getting charset from file
            # if nothing is found the default one will be used
            lines = source.split('\n')
            for l in lines:
                if 'charset' in l:

                    l = l.split('=',1)
                    charset = l[1]
                    self.set_charset(charset)
                    break

            # Convert from charset to unicode
            source = unicode(source, self._charset)
            
        # catch page elts
        lines = source.split('\n')
        for l in lines:
            if l != '':
                elt = l.split('=',1)
                elts[elt[0]] = elt[1]

        self._update_page_elts(elts)

    def _update_page_elts(self, elts):
        """
        Update the self.page_elts with new elts dict if the page content is
        newer; if so:
        -Change the content of the page
        -Make an history entry for this edition.
        -Update time edit
        -Set the author of this edition (default='')
        -Increment rev number

        Used also when :
        -Change page's attribute
        -Change page's charset

        @type   elts: dict or odict
        @param  elts: new datas for the page
        
        @rtype  Boolean
        @rparam To know if write operation is required
        """

        # Set default blank page
        tmp_elts = self.__get_blank_page_elts() # buffer for self._page_elts
        attr_has_changed = False
        txt_has_changed = False
        charset_has_changed = False
        first_init = False

        # Existing page overwrite the default one
        if self._page_elts is None:
            self._page_elts = odict()
            first_init = True
        else:
            for k, v in self._page_elts.iteritems():
                tmp_elts[k] = v

        # New datas overwrite the existing page
        for k, v in elts.iteritems():
            tmp_elts[k] = v
        
        # Check if charset has changed
        if tmp_elts['charset'] != self._charset:
            self.set_charset(tmp_elts['charset'])
            charset_has_changed = True

        # Check if attributes have changed
        keys = ['passwdread', 'passwdedit', 'passwdupload', 'passwdattr']
        for key in keys:
            
            if (key in tmp_elts and not key in self._page_elts \
                and tmp_elts[key] is not None) \
                or \
                (key in tmp_elts and key in self._page_elts \
                and tmp_elts[key] != self._page_elts[key]):

                attr_has_changed= True
                tmp_elts['rev'] = str(int(tmp_elts['rev'])+1)
                break
        
        # Check if the text page has changed
        old_txt = '' #It's a new page
        if 'text' in self._page_elts:
            old_txt = self._page_elts['text'] #It's an existing page
        new_txt = tmp_elts['text']


        if (old_txt != new_txt and first_init is False) \
        or int(tmp_elts['rev'])==0:
            txt_has_changed = True
        # else NO NEED HISTORY !

        if txt_has_changed and not attr_has_changed:

            now = str(int(time()))
            if 'time' in tmp_elts:
                time_diff = tmp_elts['time'] # time of last editing

                # If several editions ( typacally fast ones by script )
                # occure in less a second, then several revisions share the
                # same history time. These are never endless overwritten by
                # the first history until a new full second is spent
                if time_diff >= now:
                    now = str(int(tmp_elts['time'])+1)
                    logging.warning(
                    '{0} has been written with a futur rev date: +{1}s'.format(
                        self._name,
                        int(now)-time())
                    )

            # update page elts for writing
            if int(tmp_elts['rev'])==0: # Creation
                tmp_elts['ctime'] = now
                time_diff = now # else it would be None

            tmp_elts['time'] = now
            tmp_elts['rev'] = str(int(tmp_elts['rev'])+1)
            tmp_elts['agent'] = 'pmwikilib-{0}'.format(__version__)
            if not 'author' in elts: # or elts['author'] is None:
                tmp_elts['author'] = ''

            # HISTORY---------------------
            # Split elements'page & history stuff
            buf_elts = odict() # to store page's element
            history = odict()  # to store all previous history rev
            for k, v in tmp_elts.iteritems():
                if ':' in k: # we know it's an history elt with the ':' in key
                    history[k] = v
                else:
                    buf_elts[k] = v
            tmp_elts = buf_elts

            # Make and add history for this edition
            tmp_elts['author:{0}'.format(now)] = tmp_elts['author']
            tmp_elts['diff:{0}:{1}:'.format(now, time_diff)] \
                = txt2src(diff.diff(src2txt(new_txt), src2txt(old_txt)))
            tmp_elts['host:{0}'.format(now)] = tmp_elts['host']

            # Reinject  history of previous editions at the end
            for k, v in history.iteritems():
                tmp_elts[k] = v
            #---------------------------------

        # Flush all
        self._page_elts = odict() # reinit to always have the right elts order
        for k, v in tmp_elts.iteritems():
            if v is not None:
                self._page_elts[k] = v
        
        return txt_has_changed or attr_has_changed or charset_has_changed

    def __get_blank_page_elts(self):
        """
        Return a default template for self.page_elts 
        """
        blank_page_elts = odict([
            ('version', 'pmwiki-2.?.? ordered=1 urlencoded=1'),
            ('agent', 'pmwikilib-{0}'.format(__version__)),
            ('author', None),
            ('charset', self._charset),
            ('csum', None),
            ('ctime', None),
            ('description', None),
            ('host', '127.0.0.1'),
            ('name', fmt_pageName(self._name)),
            ('passwdattr', None),
            ('passwdedit', None),
            ('passwdread', None),
            ('passwdupload', None),
            ('rev', '0'),
            ('targets', None),
            ('text', None),
            ('time', None),
            ('title', None),
            ('newline', None),
            ('updatedto', None)
        ])
        return blank_page_elts

    def _write_to_file(self):
        """
        Flush self._page_elts to the file with right encoding
        """
        full_path = p.join(self._path, self._name)

        #self._page_elts['charset'] = self._charset
        txt = u'\n'.join(
                [u'{0}={1}'.format(k,v) for k, v in self._page_elts.iteritems()]
              )
        txt = unicode(txt).encode(self._charset)

        file_towrite = self._try_fileaccess(full_path, 'w')
        file_towrite.write(txt)
        file_towrite.close()

    def check_path(self):
        """
        Check if the path is reacheable
        
        @rtype    Boolean
        @rparam   True if we can reach the path else raise an Error
        """
        if not p.exists(self._path):
            raise NoValidReachPointError(
                '{0} doesn\'t exist'.format(self._path)
            )

        if not access(self._path, W_OK):
            raise NoValidReachPointError(
                '{0} is write protected'.format(self._path)
            )

        return True

    @for_human
    def read(self, author=None, password=None):
        """
        Read the page
        
        @type   author: string
        @param  author: Never used here

        @type   password: String
        @param  password: Never used here 
                          (because we can do what we want on a local file)

        @rtype  String
        @rparam Human readable content of the Pmwiki page file
        """
        full_path = p.join(self._path, self._name)
        file_toread = self._try_fileaccess(full_path, 'r')
        self._parse_file( file_toread.read() )
        file_toread.close()

        return self._page_elts['text']

    def write(self, text=None, author=None, password=None, charset=None):
        """
        Write new page's content. 
        
        If only charset is changed, the file is just written with the 
        new charset without making a new history entry.
        
        @type   text: String
        @param  text: Pmwiki's syntax text

        @type   author: string
        @param  author: Set the author name of the edition

        @type   password: String
        @param  password: Never used here
                          (because we can do what we want on a local file)

        @type   charset: String
        @param  charset: character encoding of the page.

        """
 
        # If the file existe we pre-fill the self.page_elts; so we keep previous
        # history information, ctime ...
        if p.exists(p.join(self._path, self._name)) and self._page_elts is None:
            self.read(author, password) # if we don't do, file is overwritten

        elts = {'text':txt2src(text)}
        if author is not None:
            elts['author'] = author

        if charset is not None:
            elts['charset'] = charset

        # Write only if content has changed 
        # or need re-encoding (detect via charset change)
        if self._update_page_elts(elts): 
            self._write_to_file()

    def change_attr(self, author=None, password=None, **attributes):
        """
        Change attributes'page to protect or unprotect the page with reading,
        writting, uploading and changing attributes.

        Add an history entry

        @type   author: string
        @param  author: author name for this edition

        @type   password: String
        @param  password: useless here, we dont'need it for file access

        @type   attributes: dict
        @param  attribute: key elts in dict can be 'passwdedit', 'passwdread', 
                           'passwdupload' and 'passwdattr'
        """

        if p.exists(p.join(self._path, self._name)): # page must exist
            if self._page_elts is None:
                # To avoid overwritting
                self.read(author, password)
            
            elts = {}

            for key, value in attributes.iteritems():
                if value[:1] == '@':
                    elts[key]=value
                elif value == 'clear':
                    elts[key]=None
                else:
                    # check if pass exists
                    if key in self._page_elts \
                        and check_password(value, self._page_elts[key]):
                          elts[key]=self._page_elts[key]
                    else:
                        elts[key] = encrypt(value)
            
            if self._update_page_elts(elts): #write only if content has changed
                self._write_to_file()
        else:
            raise NonExistingPageError(
                'Trying to change attributes of {0}'.format(self._name)
                )

class PmwikiPage(object):
    """
    Public API to use with page for PmWiki
    ( see: http://www.pmwiki.org/ )

    You can create, read, write and protect a page to a Pmwiki site web 
    via http request or directly to a local file.
    """

    def __init__(self, path, name):
        """
        @type   path: string
        @param  path: the way to reach the page. If start with 'http://' we 're
                        looking for an URL else for a path on local file system

        @type   name: string
        @param  name: Name formatted as 'GroupName.PageName'
        """
        # Add a trailing / if not already done
        if path[-1] != '/':
            path += '/'
        self.__path = path
        self.__name = name
        self.__page = self.__get_typePage( self.__path )
        self._text = None

    def __get_typePage(self, path):
        """
        Return derivated object of BasePage
        """
        # Search for an url...
        url = urlparse.urlparse(path)

        if url.scheme == 'http' and len(url.netloc) >0:
            return HttpPage(self.__path, self.__name)
        else: # ... finally it's a local file path
            return FilePage(self.__path, self.__name)
    
    def get_name(self):
        return self.__name;

    @property
    def text(self):
        """
        Encoded (with charset given) page's content. 
        Only used for input/output, we convert it to unicode for doing
        the job
        """
        return self._text

    @text.setter
    def text(self, value):
        if value is None:
            self._text = None
        else:
            self._text = unicode(value, self.__page._charset)

    @text.getter
    def text(self):
        if self._text is not None:
            return unicode(self._text).encode(self.__page._charset)
        return None

    @property
    def unicode_text(self):
        """
        Unicode page's content as it used internally.
        """
        return self._text

    @unicode_text.setter
    def unicode_text(self, value):
        if isinstance(value, unicode):
            self._text = value
        elif  value is None:
            self._text = None

    @unicode_text.getter
    def unicode_text(self):
        return self._text

    def check_path(self):
        """
        Check if we can reach the page.
        
        For web page: Test http connexion and try to guess if the given url is 
        a pmwiki web site.
        
        For file page: test if path exist and if our rights allow us I/O access

        @rtype  Boolean
        @rparam Indicate if we can access the page.
        """
        return self.__page.check_path()

    def read(self, author=None, password=None):
        """
        Return encoded page's content

        Author & password are required for web protected page.

        @type   author: string
        @param  author: author (as $authid) required when page is write 
                        protected and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is write protected

        @rtype  string
        @rparam Encoded content's of an existing page
        """
        self._text = self.__page.read(author, password)
        return self.text # convert to page charset via text.getter
    
    def write_text(self, text=None, author=None, password=None, charset=None):
        """
        Writting some text to the page.
        
        Author & password are required for web protected page. Also author is
        used as the author's name of an edition for both web and local file 
        page.

        @type   author: string
        @param  author: author (as $authid) required when page is write 
                        protected and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is write protected

        @type   charset: String
        @param  charset: Character encodind for the page. If not set UTF-8 will
                         be used.
        """
        #if charset is None:
            #charset = self.__page._charset
   
        if text is not None and len(text) > 0:
            self.text = text # Convert to unicode via text.setter
            self.__page.write(self._text, author, password, charset)
        else:
            logging.warning(
            '{0}: nothing to write. Page is unchanged.'.format(self.__name))

    def write(self, author=None, password=None, charset=None):
        """
        Flush the self.text into the page

        Author & password are required for web protected page. Also author is
        used as the author's name of an edition for both web and local file 
        page.

        @type   author: string
        @param  author: author (as $authid) required when page is write 
                        protected and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is write protected

        @type   charset: String
        @param  charset: Character encodind for the page. If not set UTF-8 will
                         be used.
        """
        if self._text is not None and len(self._text) > 0:
            self.__page.write(self._text, author, password, charset)
        else:
            logging.warning(
            '{0}: nothing to write. Page is unchanged.'.format(self.__name))

    def change_attr(self, author=None, password=None, **attributes):
        """
        Change attributes'page to protect or unprotect the page with reading,
        writting, uploading and changing attributes. Change attributes will add
        an history entry for both web and local file page.

        Author & password are required for web protected page. Also author is
        used as the author's name of an edition for both web and local file 
        page.

        @type   author: string
        @param  author: author (as $authid) required when page is write 
                        protected and AuthUser is enable

        @type   password: String
        @param  password: password needed by an authentication request when page
                          is write protected

        @type   attributes: dict
        @param  attribute: key elts in dict can be 'passwdedit', 'passwdread', 
                           'passwdupload' and 'passwdattr'
        """
        good_attrs = {'passwdread':None, 'passwdedit':None, 'passwdattr':None,
                     'passwdupload':None}
        for k in good_attrs.keys():
            if k in attributes and len(attributes[k]) > 0:
                good_attrs[k] = attributes[k]
            else:
                del good_attrs[k]
        if len(good_attrs) == 0:
            logging.warning('{0}: bad attrs({i}). Page is unchanged.'.format(
                self.__name, ''.join([at for at in attributes.keys()]))
            )
        else:
            self.__page.change_attr( author, password, **good_attrs )

    def set_charset(self, charset):
        """
        Change internal parameter character encoding of the page.

        Use to convert to unicode when setting the text property and to encode 
        page's content when writting if charset is not given with write or
        write_text method.

        Allow to re-encode page if called just before write or write_text. 
        
        Note : If only charset is changed before write operation, an history 
        entry will be added for web page but not local file.
        """
        self.__page.set_charset(charset)

